local grabbableitems = {
	[MT_POPCORN] = true,
	[MT_SODA] = true,
}


local function buttonPressed(p, bt)
	return (p.cmd.buttons & bt and not (p.theater.oldbuttons & bt))
end

local function findAimedMobj(p)
	local pmo = p.mo
	local pangle = pmo.angle

	local maxdist = 128 * FU
	local maxheight = 64 * FU
	local maxangle = ANGLE_45

	local x, y, z = pmo.x, pmo.y, pmo.z
	local x1, y1, z1 = x - maxdist, y - maxdist, z - maxheight
	local x2, y2, z2 = x + maxdist, y + maxdist, z + maxheight
	local bestdelta, found = INT32_MAX, nil

	-- Check for mobjs around the player
	for mo in mobjs.iterate() do
		if mo.x > x1 and mo.x < x2
		and mo.y > y1 and mo.y < y2
		and mo.z > z1 and mo.z < z2 then
			local dist = R_PointToDist2(x, y, mo.x, mo.y)

			if dist < maxdist
			and grabbableitems[mo.type]
			and mo ~= p.theater.lefthanditem and mo ~= p.theater.righthanditem
			and mo ~= pmo then
				local angle = abs(pangle - R_PointToAngle2(x, y, mo.x, mo.y))
				if angle < maxangle then
					-- Use some basic heuristic to determine if that mobj
					-- if more likely to be the one the player is thinking of
					local delta = FixedDiv(angle, maxangle) / 2
					            + FixedDiv(dist, maxdist)
					            + FixedDiv(abs(z - mo.z), maxheight)

					if delta < bestdelta then
						bestdelta, found = delta, mo
					end
				end
			end
		end
	end

	return found
end

local function tryGrabHandItem(p, grounditem)
	local pmo = p.mo

	local slot
	if not p.theater.righthanditem then
		slot = "righthanditem"
	elseif not p.theater.lefthanditem then
		slot = "lefthanditem"
	else
		return
	end

	local handitem
	if grounditem.spawnpoint then
		handitem = P_SpawnMobj(pmo.x, pmo.y, pmo.z, grounditem.type)
	else
		handitem = grounditem
	end

	p.theater[slot] = handitem
	handitem.target = pmo

	p.theater[slot].flags = $ | MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
end

local function eatPopcorn(p, item, slot)
	if item.theater_quantity == 0 then
		P_RemoveMobj(item)
		p.theater[slot] = nil
	else
		item.theater_quantity = $ - 1

		if item.theater_quantity == 0 then
			item.frame = C
		elseif item.theater_quantity <= 5 then
			item.frame = B
		end

		-- Spawn crumbs
		for _ = 1, P_RandomRange(1, 10) do
			local crumb = P_SpawnMobj(item.x, item.y, item.z + item.height, MT_POPCORNCRUMB)
			crumb.scale = item.scale
			crumb.fuse = 3 * 60 * TICRATE

			local angle = P_RandomRange(0, 65535) * 65536
			local speed = P_RandomRange(0, 4)
			crumb.momx = p.mo.momx + speed * cos(angle)
			crumb.momy = p.mo.momy + speed * sin(angle)
			crumb.momz = p.mo.momz + P_RandomRange(1, 5) * item.scale
		end

		S_StartSound(p.mo, sfx_popeat)
	end
end

local function drinkSoda(p, item, slot)
	if item.theater_quantity == 0 then
		P_RemoveMobj(item)
		p.theater[slot] = nil
	else
		item.theater_quantity = $ - 1
		S_StartSound(p.mo, item.theater_quantity == 0 and sfx_sip2 or sfx_sip1)
	end
end

local function tryUseHandItem(p)
	if not (p.theater.lefthanditem or p.theater.righthanditem) then return end

	local slot, item
	repeat
		if P_RandomChance(FU / 2) then
			slot = "lefthanditem"
		else
			slot = "righthanditem"
		end

		item = p.theater[slot]
	until item

	if not item then return end

	if item.type == MT_POPCORN then
		eatPopcorn(p, item, slot)
	elseif item.type == MT_SODA then
		drinkSoda(p, item, slot)
	end
end

function theater.updateGrabbing(p)
	local pmo = p.mo
	if not pmo then return end

	local item = p.theater.lefthanditem
	if item and (not item.valid or item.target ~= pmo) then
		p.theater.lefthanditem = nil
	end
	local item = p.theater.righthanditem
	if item and (not item.valid or item.target ~= pmo) then
		p.theater.righthanditem = nil
	end

	if buttonPressed(p, BT_TOSSFLAG) then
		local item = findAimedMobj(p)

		if item then
			tryGrabHandItem(p, item)
		else
			tryUseHandItem(p)
		end
	end

	-- Update the position of the carried items
	local leftitem = p.theater.lefthanditem
	if leftitem then
		local dist = pmo.radius * 2
		local angle = pmo.angle + 8192 * FU
		local x = pmo.x + FixedMul(dist, cos(angle))
		local y = pmo.y + FixedMul(dist, sin(angle))
		local z = pmo.z + pmo.height / 2
		P_TeleportMove(leftitem, x, y, z)
	end

	local rightitem = p.theater.righthanditem
	if rightitem then
		local dist = pmo.radius * 2
		local angle = pmo.angle - 8192 * FU
		local x = pmo.x + FixedMul(dist, cos(angle))
		local y = pmo.y + FixedMul(dist, sin(angle))
		local z = pmo.z + pmo.height / 2
		P_TeleportMove(rightitem, x, y, z)
	end
end


addHook("PlayerQuit", function(p)
	if p.theater then
		local item = p.theater.lefthanditem
		if item and item.valid then
			item.target = nil
			item.flags = $ & ~MF_NOGRAVITY
		end

		local item = p.theater.righthanditem
		if item and item.valid then
			item.target = nil
			item.flags = $ & ~(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
		end
	end
end)
