rawset(_G, "theater", {})


local joined = false

theater.moviedefs = {}


local cv_allowmovieseek = CV_RegisterVar{"allowmovieseek", "Off", CV_NETVAR, CV_OnOff}


local function playMovieMusic()
	local def = theater.moviedefs[theater.movie.id]

	S_ChangeMusic(def.music)

	S_SetMusicPosition(theater.movie.pos * 1000 / TICRATE)

	local p = consoleplayer
	if p and p.theater then
		S_SetInternalMusicVolume(p.theater.volume)
	end
end

local function applyMovieState()
	local def = theater.moviedefs[theater.movie.id]

	local frame = FixedMul(def.frames, FixedDiv(theater.movie.pos, def.duration))
	frame = min($, def.frames - 1)

	-- Update screen texture
	local texfmt = def.textures
	local numdigits = #texfmt - texfmt:find("%**$") + 1
	local prefix = texfmt:sub(1, #texfmt - numdigits)
	local suffix = string.format("%0" .. numdigits .. "d", frame)
	theater.movie.screen.midtexture = R_TextureNumForName(prefix .. suffix)

	-- Ensure audio is in sync
	local audiopos = theater.movie.pos * 1000 / TICRATE
	if leveltime % TICRATE == 0 -- Check every second
	and abs(S_GetMusicPosition() - audiopos) > 100 then
		S_SetMusicPosition(audiopos)
	end
end

local function updateMovie()
	local def = theater.moviedefs[theater.movie.id]

	-- Increment pos
	if theater.movieEnded() then
		return
	else
		theater.movie.pos = min($ + 1, def.duration - 1)

		if theater.movieEnded() then
			-- Make screen black
			theater.movie.screen.midtexture = R_TextureNumForName("~031")
			S_StopMusic()
			return
		end
	end

	applyMovieState()
end

function theater.movieEnded()
	local def = theater.moviedefs[theater.movie.id]
	return (theater.movie.pos >= def.duration - 1)
end

function theater.playMovie(id, theatertag)
	theatertag = $ or 10000 -- !!!

	local def = theater.moviedefs[id]

    theater.movie = {}

	theater.movie.id = id
	theater.movie.pos = 0

	local screentag
	if def.width == 120 then
		screentag = theatertag + 2
	elseif def.width == 160 then
		screentag = theatertag + 4
	elseif def.width == 180 then
		screentag = theatertag + 8
	elseif def.width == 200 then
		screentag = theatertag + 6
	else
		error("movie width not supported.")
	end
	theater.movie.screen = lines[P_FindSpecialLineFromTag(0, screentag)].frontside

	for s in sectors.iterate do
		if s.tag == theatertag then
			s.floorheight = (s.ceilingheight + s.floorheight) / 2
			s.floorheight = $ - def.height * FU / 2
			s.ceilingheight = s.floorheight + (def.height - 1) * FU
		end
	end

	for l in lines.iterate do
		if l.tag >= theatertag and l.tag <= theatertag + 6 then
			local side = l.frontside
			if l.tag == theatertag + 1 then
				side.midtexture = R_CheckTextureNumForName("~255")
			elseif l.tag == screentag + 1 then
				side.midtexture = R_CheckTextureNumForName("~031")
			else
				side.midtexture = -1
			end
		end
	end

	playMovieMusic()
end

function theater.addMovie(moviedef)
	theater.moviedefs[#theater.moviedefs + 1] = moviedef
	theater.moviedefs[moviedef.id] = moviedef
end

addHook("LinedefExecute", function(l)
	if not theater.movie or theater.movieEnded() then
		--theater.playMovie("pokemonsoccer", l.tag)
	end
end, "THEATERPLAY")

addHook("ThinkFrame", function()
	for p in players.iterate do
		if not p.theater then
			p.theater = {}
			p.theater.volume = 5
		end
	end

	if joined and consoleplayer then
		joined = false

		if theater.movie then
			playMovieMusic()
		end
	end

	if theater.movie and theater.movie.screen.valid then
		updateMovie()
	end

	theater.checkCrumbs()
end)

addHook("PlayerThink", function(p)
	if not p.theater then return end

	theater.updateGrabbing(p)

	p.theater.oldbuttons = p.cmd.buttons
end)

addHook("LinedefExecute", function(l, mo)
	if not mo then return end
	local p = mo.player
	if not (p and p.theater) then return end

	p.theater.volume = l.frontside.textureoffset / FU

	if p == consoleplayer then
		S_SetInternalMusicVolume(p.theater.volume)
	end
end, "THEATERVOLUME")

COM_AddCommand("theaterplay", function(p, id)
	if not id then
		CONS_Printf(p, "Usage: theaterplay <movie ID>")

		CONS_Printf(p, "Available IDs:")
		for _, def in ipairs(theater.moviedefs) do
			CONS_Printf(p, "  - \x82" .. def.id .. "\x80: " .. def.name)
		end
	elseif not theater.moviedefs[id] then
		CONS_Printf(p, 'Unknown movie "' .. id .. '".')
	else
		theater.playMovie(id)
	end
end, COM_ADMIN)

COM_AddCommand("theaterframe", function(p, pos)
	if not theater.movie then return end

	local def = theater.moviedefs[theater.movie.id]

	pos = tonumber($)
	if pos ~= nil then
		if not cv_allowmovieseek.value
		and not (p == server or IsPlayerAdmin(p)) then
			CONS_Printf(p, "Seeking is currently disabled.")
			return
		end

		--theater.movie.pos = min(max(pos * TICRATE, 0), def.duration - 1)
		theater.movie.pos = min(max(pos, 0), def.duration - 1)

		local audiopos = theater.movie.pos * 1000 / TICRATE
		playMovieMusic()
		S_SetMusicPosition(audiopos)
	else
		CONS_Printf(p, theater.movie.pos .. "/" .. def.duration)
	end
end)

addHook("MapChange", function()
	theater.movie = nil
end)

addHook("NetVars", function(net)
	if not isserver then
		joined = true
	end

	theater.movie = net($)
end)
